@@ -15,18 +15,22 @@
 #
 # END HEADER
 
-from __future__ import division, print_function, absolute_import
+from __future__ import absolute_import, division, print_function
 
 import os
 import sys
-
-import coverage
+import traceback
+from inspect import getframeinfo
 
 import hypothesis
-from hypothesis.errors import StopTest, DeadlineExceeded, \
-    HypothesisException, UnsatisfiedAssumption
-from hypothesis.internal.compat import text_type, binary_type, \
-    encoded_filepath
+from hypothesis.errors import (
+    DeadlineExceeded,
+    HypothesisException,
+    MultipleFailures,
+    StopTest,
+    UnsatisfiedAssumption,
+)
+from hypothesis.internal.compat import binary_type, encoded_filepath, text_type
 
 if False:
     from typing import Dict  # noqa
@@ -47,21 +51,19 @@ def belongs_to(package):
         cache[ftype][filepath] = result
         cache[type(new_filepath)][new_filepath] = result
         return result
-    accept.__name__ = 'is_%s_file' % (package.__name__,)
+
+    accept.__name__ = "is_%s_file" % (package.__name__,)
     return accept
 
 
-PREVENT_ESCALATION = os.getenv('HYPOTHESIS_DO_NOT_ESCALATE') == 'true'
+PREVENT_ESCALATION = os.getenv("HYPOTHESIS_DO_NOT_ESCALATE") == "true"
 
 FILE_CACHE = {}  # type: Dict[bytes, bool]
 
 
 is_hypothesis_file = belongs_to(hypothesis)
-is_coverage_file = belongs_to(coverage)
 
-HYPOTHESIS_CONTROL_EXCEPTIONS = (
-    DeadlineExceeded, StopTest, UnsatisfiedAssumption
-)
+HYPOTHESIS_CONTROL_EXCEPTIONS = (DeadlineExceeded, StopTest, UnsatisfiedAssumption)
 
 
 def mark_for_escalation(e):
@@ -73,16 +75,31 @@ def escalate_hypothesis_internal_error():
     if PREVENT_ESCALATION:
         return
     error_type, e, tb = sys.exc_info()
-    if getattr(e, 'hypothesis_internal_always_escalate', False):
+    if getattr(e, "hypothesis_internal_always_escalate", False):
         raise
-    import traceback
     filepath = traceback.extract_tb(tb)[-1][0]
     if is_hypothesis_file(filepath) and not isinstance(
-        e, (HypothesisException,) + HYPOTHESIS_CONTROL_EXCEPTIONS,
+        e, (HypothesisException,) + HYPOTHESIS_CONTROL_EXCEPTIONS
     ):
         raise
-    # This is so that if we do something wrong and trigger an internal Coverage
-    # error we don't try to catch it. It should be impossible to trigger, but
-    # you never know.
-    if is_coverage_file(filepath):  # pragma: no cover
-        raise
+
+
+def get_trimmed_traceback():
+    """Return the current traceback, minus any frames added by Hypothesis."""
+    error_type, _, tb = sys.exc_info()
+    # Avoid trimming the traceback if we're in verbose mode, or the error
+    # was raised inside Hypothesis (and is not a MultipleFailures)
+    if hypothesis.settings.default.verbosity >= hypothesis.Verbosity.debug or (
+        is_hypothesis_file(traceback.extract_tb(tb)[-1][0])
+        and not isinstance(error_type, MultipleFailures)
+    ):
+        return tb
+    while tb is not None and (
+        # If the frame is from one of our files, it's been added by Hypothesis.
+        is_hypothesis_file(getframeinfo(tb.tb_frame)[0])
+        # But our `@proxies` decorator overrides the source location,
+        # so we check for an attribute it injects into the frame too.
+        or tb.tb_frame.f_globals.get("__hypothesistracebackhide__") is True
+    ):
+        tb = tb.tb_next
+    return tb
